#include <glib.h>
#include <gtk/gtk.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include "fftv.h"
#include "tv-freq.h"
#include "tv-win.h"

long current_time();
static GtkWidget *spinner_minutes_left;
static GtkWidget *tv_countdown_window = NULL, *entry_shutdown;
void align_with_ui_window(GtkWidget *win);
static char *shutdown_cmd;
static long tv_countdown_start_time, tv_countdown_finish_time;

static int tag_tv_countdown;
static int tv_countdown_minutes;
int b_do_exit;


static char *get_shutdown_exec()
{
  static char shutdown_exec[256];
  char fftv_states_dir[256];

  get_fftv_states_dir(fftv_states_dir);

  sprintf(shutdown_exec, "%s/fftv-shutdown-exec", fftv_states_dir);
  return shutdown_exec;
}


void load_shutdown_conf()
{

  if (shutdown_cmd)
    return;

  char *shutdown_exec = get_shutdown_exec();

  FILE *fp;
  if ((fp=fopen(shutdown_exec,"r"))==NULL) {
    shutdown_cmd = "halt";
    return;
  }

  char buf[256];

  buf[0]=0;
  fgets(buf,sizeof(buf),fp);

  if (buf[strlen(buf)-1]=='\n')
      buf[strlen(buf)-1]=0;

  shutdown_cmd = strdup(buf);

  fclose(fp);
}


static void close_tv_countdown_window()
{
  gtk_widget_destroy(tv_countdown_window);
  tv_countdown_window = NULL;
}

static void stop_countdown()
{
  if (tag_tv_countdown)
    gtk_timeout_remove(tag_tv_countdown);
  tag_tv_countdown = 0;
  tv_countdown_start_time = tv_countdown_finish_time = 0;
}


void exec_shutdown_cmd()
{
  if (!tv_countdown_finish_time)
    return;

  system(shutdown_cmd);
}



static gboolean timeout_tv_countdown()
{
  dbg("time is up\n");
  b_do_exit = 1;

  exec_shutdown_cmd();

  return FALSE;
}


static gboolean close_application( GtkWidget *widget,
                                   GdkEvent  *event,
                                   gpointer   data )
{
  close_tv_countdown_window();
  return FALSE;
}



static gboolean cb_start( GtkWidget *widget, GdkEvent  *event, gpointer   data )
{
  stop_countdown();
  tv_countdown_minutes = (int) gtk_spin_button_get_value(GTK_SPIN_BUTTON(spinner_minutes_left));

  tag_tv_countdown=gtk_timeout_add(tv_countdown_minutes * 60 * 1000,
                                   timeout_tv_countdown,NULL);

  tv_countdown_start_time = current_time();
  tv_countdown_finish_time = tv_countdown_start_time + tv_countdown_minutes * 60;


  const char *text_shutdown = gtk_entry_get_text(GTK_ENTRY(entry_shutdown));

  shutdown_cmd = strdup(text_shutdown);

  char *shutdown_exec = get_shutdown_exec();

  FILE *fp;
  if ((fp=fopen(shutdown_exec,"w"))) {
    fprintf(fp,"%s\n", text_shutdown);
    fclose(fp);
  }

  close_tv_countdown_window();

  return TRUE;
}

static gboolean cb_stop( GtkWidget *widget, GdkEvent  *event, gpointer   data )
{
  stop_countdown();

  close_tv_countdown_window();
  return TRUE;
}

void cb_spinner_value_changed_minutes_left(GtkSpinButton *spinbutton,
                                      gpointer user_data)
{
#if 0
  int value = (int) gtk_spin_button_get_value(spinbutton);

  dbg("x:%d\n", value);
#endif
}



void align_with_ui_window(GtkWidget *win);

void create_tv_countdown_window()
{
  GtkWidget *box_top;

  if (tv_countdown_window) {
    gtk_window_present(GTK_WINDOW(tv_countdown_window));
    return;
  }

  tv_countdown_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  align_with_ui_window(tv_countdown_window);
  gtk_container_set_border_width (GTK_CONTAINER (tv_countdown_window), 5);

  gtk_window_set_title(GTK_WINDOW(tv_countdown_window), "fftv countdown shutdown");

  g_signal_connect (G_OBJECT (tv_countdown_window), "delete_event",
                      G_CALLBACK (close_application),
                      NULL);

  box_top = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (tv_countdown_window), box_top);

  if (tv_countdown_finish_time) {
    tv_countdown_minutes = (tv_countdown_finish_time - current_time()) / 60.0 + 0.5;
  } else {
    time_t ltime=time(NULL);
    struct tm ltm;

    localtime_r(&ltime, &ltm);
    tv_countdown_minutes = 60 - (ltm.tm_min % 60);
  }


  GtkWidget *hbox_tv_minutes_left = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (box_top), hbox_tv_minutes_left, FALSE, FALSE, 5);
  GtkWidget *label_tv_minutes_left = gtk_label_new("Minutes left");
  gtk_box_pack_start (GTK_BOX (hbox_tv_minutes_left), label_tv_minutes_left, FALSE, FALSE, 0);

  double step = 1.0;
  GtkAdjustment *adj_tv_minutes_left = (GtkAdjustment *) gtk_adjustment_new (
       tv_countdown_minutes, 0.0, 640.0, step, step, step);
  spinner_minutes_left = gtk_spin_button_new (adj_tv_minutes_left, step, 0);

  g_signal_connect (G_OBJECT(spinner_minutes_left), "value-changed",
                    G_CALLBACK(cb_spinner_value_changed_minutes_left), NULL);

  gtk_widget_set_size_request (spinner_minutes_left, 60, -1);
  gtk_box_pack_start (GTK_BOX (hbox_tv_minutes_left), spinner_minutes_left, FALSE, FALSE, 0);


  GtkWidget *hbox_shutdown_exec = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (box_top), hbox_shutdown_exec, FALSE, FALSE, 5);
  GtkWidget *label_shutdown_exec = gtk_label_new("Shutdown exec");
  gtk_box_pack_start (GTK_BOX (hbox_shutdown_exec), label_shutdown_exec, FALSE, FALSE, 0);

  load_shutdown_conf();

  entry_shutdown = gtk_entry_new ();
  gtk_box_pack_start (GTK_BOX (hbox_shutdown_exec), entry_shutdown, FALSE, TRUE, 10);
  gtk_entry_set_max_length (GTK_ENTRY (entry_shutdown), 128);
  gtk_entry_set_width_chars(GTK_ENTRY (entry_shutdown), 15);
  gtk_entry_set_text (GTK_ENTRY (entry_shutdown), shutdown_cmd);
//    gtk_widget_set_size_request(entry, 100,24);


  GtkWidget *hbox_start_stop = gtk_hbox_new (TRUE, 0);
  gtk_box_pack_start (GTK_BOX (box_top), hbox_start_stop, FALSE, FALSE, 5);

  GtkWidget *button_start = gtk_button_new_from_stock(GTK_STOCK_EXECUTE);
  g_signal_connect_swapped (G_OBJECT (button_start), "clicked",
                            G_CALLBACK (cb_start),
                            G_OBJECT (tv_countdown_window));
  gtk_box_pack_start (GTK_BOX (hbox_start_stop), button_start, TRUE, TRUE, 10);

  GtkWidget *button_stop = gtk_button_new_from_stock(GTK_STOCK_STOP);
  g_signal_connect_swapped (G_OBJECT (button_stop), "clicked",
                            G_CALLBACK (cb_stop),
                            G_OBJECT (tv_countdown_window));
  gtk_box_pack_start (GTK_BOX (hbox_start_stop), button_stop, TRUE, TRUE, 10);

  GTK_WIDGET_SET_FLAGS (button_start, GTK_CAN_DEFAULT);
  gtk_widget_show_all (tv_countdown_window);
}


