#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>
#include <sys/time.h>
#include <time.h>

#include <gtk/gtk.h>

#include "reminder.h"

#include <string.h>
#include <ctype.h>

#include "tv-freq.h"
#include "fftv.h"
#include "xkey.h"
#include "vidstr.h"
#include "tv-win.h"

extern gboolean b_ch_monitor;
void save_current_channel_name();
void disp_rec_status();
void disp_rec_output_status();
void set_radio_record_active_image();
void set_radio_pause_active_image();
void set_radio_pause_invactive_image();
void set_radio_record_inactive_image();
void set_radio_pause_inactive_image();
void set_tv_record_active_image();
void set_tv_record_inactive_image();
void set_tv_pause_active_image();
void set_tv_pause_inactive_image();
void toggle_channel_monitor();


void start_recording()
{
  if (record_status == RECORD_ON) {
    if (fftv_mode & FFTV_MODE_TV)
      disp_rec_output_status();
    return;
  }


  set_record_status(RECORD_ON);

  if (fftv_mode & FFTV_MODE_TV) {
    if (TV_WIN_DISP_MODE_GRAB_RECORD) {
      set_tv_win_refresh_mode(TV_WIN_REFRESH_GRAB);
      if (!TV_WIN_DISP_MODE_GRAB)
        init_v4l_mbuf();
    }
    else {
      set_tv_win_refresh_mode(TV_WIN_REFRESH_OVERLAY);
      init_v4l_mbuf();
    }

    if (b_ch_monitor)
      toggle_channel_monitor();

    set_record_status(RECORD_ON);

    set_fftv_cmd_state("Record");
    disp_rec_status();

    set_tv_record_active_image();
    set_tv_pause_inactive_image();

    if (TV_WIN_DISP_MODE_GRAB && TV_WIN_DISP_MODE_GRAB_RECORD) {
    }
    else
    if (!TV_WIN_DISP_MODE_GRAB && TV_WIN_DISP_MODE_GRAB_RECORD) {
      tv_pic_off();
      create_x_shm_image();
    } else
    if (TV_WIN_DISP_MODE_GRAB && !TV_WIN_DISP_MODE_GRAB_RECORD) {
      tv_pic_on();
    }

    fftv_read_frame(0); // drop the first frame
  } else {
    set_record_status(RECORD_ON);
    set_radio_record_active_image();
    set_radio_pause_inactive_image();

    set_fftv_cmd_state("Record-Radio");
    dbg(".... radio record\n");
  }

  save_current_channel_name();

  break_tv_disp_loop(TRUE);
}

void fftv_grab_close();

void pause_recording()
{
  disp_rec_output_status();

  if (record_status != RECORD_ON)
    return;


  if (TV_WIN_DISP_MODE_GRAB)
    set_tv_win_refresh_mode(TV_WIN_REFRESH_GRAB);
  else
    set_tv_win_refresh_mode(TV_WIN_REFRESH_OVERLAY);

  if (fftv_mode & FFTV_MODE_TV) {
    set_fftv_cmd_state("Pause");
    set_record_status(RECORD_PAUSE);
    disp_rec_status();

    set_tv_record_inactive_image();
    set_tv_pause_active_image();

    if (!TV_WIN_DISP_MODE_GRAB)
      tv_pic_on();

  } else {
    set_radio_record_inactive_image();
    set_radio_pause_active_image();
    set_fftv_cmd_state("Pause-Radio");
    set_record_status(RECORD_PAUSE);
  }
}


int b_stop_recording;

void stop_recording()
{
  if (record_status == RECORD_OFF) {
    return;
  }

  if (TV_WIN_DISP_MODE_GRAB)
    set_tv_win_refresh_mode(TV_WIN_REFRESH_GRAB);
  else
    set_tv_win_refresh_mode(TV_WIN_REFRESH_OVERLAY);

  b_stop_recording = 1;

  if (fftv_mode & FFTV_MODE_TV) {
    if (!TV_WIN_DISP_MODE_GRAB)
      fftv_grab_close();

    set_fftv_cmd_state("Normal");
    set_record_status(RECORD_STOP);
    disp_rec_status();

    set_tv_pause_inactive_image();
    set_tv_record_inactive_image();

    if (!TV_WIN_DISP_MODE_GRAB && TV_WIN_DISP_MODE_GRAB_RECORD) {
      tv_pic_on();
    }
  } else {
    set_fftv_cmd_state("FM-radio");
    set_radio_pause_inactive_image();
    set_radio_record_inactive_image();
    set_record_status(RECORD_STOP);
  }
}


void update_time_frame();
extern int b_do_exit;

void wait_for_recording()
{
  do {
    tv_disp_main_loop();

    if (b_do_exit)
      return;
  } while ((record_status==RECORD_PAUSE || record_status==RECORD_OFF));

  dbg("record_status:%d\n",record_status);

  if (!TV_WIN_DISP_MODE_GRAB && TV_WIN_DISP_MODE_GRAB_RECORD) {
    fftv_read_frame(0); // drop the first frame
  }

  update_time_frame();
}
