// tv win grab mode display

/*
    cropped image are written to XShm/Xv image directly

 */

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>

#include <linux/videodev.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <time.h>
#include <sys/shm.h>
#include <sys/poll.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>
#include <X11/extensions/XShm.h>
#include <X11/extensions/Xvlib.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <memory.h>

#include "tv-freq.h"
#include "fftv.h"
#include "vidstr.h"
#include "tv-win.h"

extern Window root;
extern GC gc;

static struct video_mmap gb_buf;
static guchar *video_buf;
static struct video_mbuf gb_buffers;
static int grab_sizeX, grab_sizeY;
static int grab_sizeX_cur, grab_sizeY_cur;
static int grab_pixN;
static int grab_bytesN;
static int gb_frame;

extern int user_grab_sizeX, user_grab_sizeY;
static int user_grab_pixN;
static char *user_grab_image;
static int user_grab_imageN;
static int x_disp_imageN;

int visible_ofs_x0, visible_ofs_x1;

static XvPortID         xvport;

extern Display *dpy;

static Visual *xvisual;
static int xdepth;
static XImage *ximage, *ximage_scale;
static XShmSegmentInfo xshminfo, xshminfo_scale;

static XvImage *xvimage;
static int xvmode;


extern Window tv_win;
extern GC gc;
void init_Xv_port();


#define PIX_BYTES_24 (3)

#define N_PIX_I420(n) ((n)*3/2)
#define N_PIX_RGB24(n) ((n)* PIX_BYTES_24)


#if 1
GRAB_FMT fftv_grab_fmt = FFTV_I420;
#else
GRAB_FMT fftv_grab_fmt = FFTV_RGB24;
#endif

void set_grab_frame_size()
{
  grab_sizeX = user_grab_sizeX + visible_ofs_x0 + visible_ofs_x1;
  grab_sizeY = user_grab_sizeY;
  grab_pixN = grab_sizeX * grab_sizeY;

  user_grab_pixN = user_grab_sizeX * user_grab_sizeY;

  if (fftv_grab_fmt == FFTV_I420) {
    grab_bytesN = N_PIX_I420(grab_pixN);
    user_grab_imageN = N_PIX_I420(user_grab_pixN);
  }
  else
  if (fftv_grab_fmt == FFTV_RGB24) {
    grab_bytesN = N_PIX_RGB24(grab_pixN);
    user_grab_imageN = N_PIX_RGB24(user_grab_pixN);
  }

  user_grab_image = realloc(user_grab_image, user_grab_imageN);
#if 0
  dbg("user_grab_sizeX:%d user_grab_sizeY:%d  ofs: %d %d   user_grab_imageN:%d grab_bytesN:%d\n",
     user_grab_sizeX, user_grab_sizeY,
     visible_ofs_x0, visible_ofs_x1, user_grab_imageN, grab_bytesN);
#endif
}

#if 1
// using a loop to find an optimal solution
// grab width of bt-878 must be multiple of 16, otherwise you will get a green screen
void calc_grab_crop_win_X(int szx, int *rofs_x0, int *rofs_x1)
{
  if (pchan_current->ofs_x0 <=0  && pchan_current->ofs_x1 <=0) {
    *rofs_x0=*rofs_x1=0;
    return;
  }

  float r_x0 = (float)pchan_current->ofs_x0 /
               (pchan_current->ofs_x0 + szx + pchan_current->ofs_x1);
  float r_x1 = (float)pchan_current->ofs_x1 /
               (pchan_current->ofs_x0 + szx + pchan_current->ofs_x1);

  int s_ox0=pchan_current->ofs_x0, s_ox1=pchan_current->ofs_x1;

  if (s_ox0 % 2)
    s_ox0++;

  if (s_ox1 % 2)
    s_ox1++;

  int m_x0=0, m_x1=0;
  float m_cost = 10000.0;

  int ox0, ox1;
  for(ox0=s_ox0; ox0 < 100; ox0+=2)
  for(ox1=s_ox1; ox1 < 100; ox1+=2) {
    if ((ox0 + ox1) % 16)
      continue;

    float r0 = (float)ox0 / (ox0+ szx + ox1);
    float r1 = (float)ox1 / (ox0+ szx + ox1);

    float cost = r0 - r_x0 + r1 - r_x1;

    if ((pchan_current->ofs_x0 > pchan_current->ofs_x1 && ox0 < ox1) ||
        (pchan_current->ofs_x0 < pchan_current->ofs_x1 && ox0 > ox1))
      cost += 100.0;

    if (cost < m_cost) {
       m_x0 = ox0;
       m_x1 = ox1;
       m_cost = cost;
    }
  }


  *rofs_x0 = m_x0;
  *rofs_x1 = m_x1;

  dbg("%d %d m_x0:%d %d\n", pchan_current->ofs_x0, pchan_current->ofs_x1, m_x0, m_x1);
}
#endif


void init_xvisual()
{
  dbg("... init_xvisual\n");
  if (tv_win_disp_mode & (TV_GRAB_XV|TV_GRAB_XV_RECORD) &&  fftv_grab_fmt == FFTV_I420) {
    init_Xv_port();

    if (fftv_grab_fmt == FFTV_I420)
      return;
  }


  GdkVisual* gvisual = gdk_visual_get_best_with_type (GDK_VISUAL_TRUE_COLOR);

  xvisual = GDK_VISUAL_XVISUAL(gvisual);
  xdepth = gvisual->depth;
  dbg("xvisual:%x  %d\n", xvisual, xdepth);
}

void close_image()
{
  XShmDetach(dpy, &xshminfo);

  if (xvimage) {
    XFree(xvimage);
    xvimage = NULL;
  } else
  if (ximage) {
    XDestroyImage(ximage);
    ximage = NULL;
  }

  shmdt(xshminfo.shmaddr);

  if (ximage_scale) {
    XShmDetach(dpy, &xshminfo_scale);
    XDestroyImage(ximage_scale);
    ximage_scale = NULL;
    shmdt(xshminfo_scale.shmaddr);
  }
}


void create_x_shm_image()
{
//  void            *orig_handler;
  int data_size, data_size_scale;

  if (fftv_grab_fmt == FFTV_I420 && !xvport)
    init_Xv_port();

  if (grab_sizeX_cur == user_grab_sizeX && grab_sizeY_cur == user_grab_sizeY)
    return;
#if 1
#define SCALE_SIZE_X (1280)
#define SCALE_SIZE_Y (1024)
#else
#define SCALE_SIZE_X (800)
#define SCALE_SIZE_Y (600)
#endif

  grab_sizeX_cur = user_grab_sizeX; grab_sizeY_cur = user_grab_sizeY;

  if (xvimage || ximage)
    close_image();

  bzero(&xshminfo, sizeof(XShmSegmentInfo));

  if (tv_win_disp_mode == TV_GRAB_MIT_SHM || (fftv_grab_fmt & (FFTV_RGB24))) {
    ximage = XShmCreateImage(dpy, xvisual, xdepth, ZPixmap, NULL, &xshminfo,
                             user_grab_sizeX, user_grab_sizeY);
    if (!ximage)
         p_err("X11 SHM error 2");

    x_disp_imageN = data_size = ximage->bytes_per_line * user_grab_sizeY;

    dbg("xdepth:%d bpl:%d ux:%d uy:%d dsz:%d\n", xdepth,
      ximage->bytes_per_line, user_grab_sizeX, user_grab_sizeY, data_size);
  }
  else {
    dbg("XvShmCreateImage: %d %d\n", user_grab_sizeX, user_grab_sizeY);
    if (!(xvimage = XvShmCreateImage(dpy, xvport, xvmode, NULL,
                                     user_grab_sizeX, user_grab_sizeY, &xshminfo)))
         p_err("X11 SHM error 2");

    x_disp_imageN = data_size = xvimage->data_size;
  }

  xshminfo.shmid = shmget(IPC_PRIVATE, data_size, IPC_CREAT | 0777);

  if (xshminfo.shmid < 0) {
    p_err("X11 SHM error 3  %d", xshminfo.shmid);
  }

  xshminfo.shmaddr = (char *) shmat(xshminfo.shmid, 0, 0);
  if ((void *)-1 == xshminfo.shmaddr)
       p_err("X11 SHM error 4");

  if (tv_win_disp_mode == TV_GRAB_MIT_SHM || (fftv_grab_fmt & (FFTV_RGB24)))
    ximage->data = xshminfo.shmaddr;
  else
    xvimage->data = xshminfo.shmaddr;

  xshminfo.readOnly = False;

  XShmAttach(dpy, &xshminfo);
  XSync(dpy, False);
  shmctl(xshminfo.shmid, IPC_RMID, 0);

  if (tv_win_disp_mode == TV_GRAB_MIT_SHM || (fftv_grab_fmt & (FFTV_RGB24))) {
    ximage_scale = XShmCreateImage(dpy, xvisual, xdepth, ZPixmap, NULL, &xshminfo_scale,
                             SCALE_SIZE_X, SCALE_SIZE_Y);
    if (!ximage_scale)
         p_err("X11 SHM error 2 ximage_scale");

    data_size_scale = ximage_scale->bytes_per_line * SCALE_SIZE_Y;

    xshminfo_scale.shmid = shmget(IPC_PRIVATE, data_size_scale, IPC_CREAT | 0777);

    if (xshminfo_scale.shmid < 0) {
         p_err("X11 SHM error 3  %d", xshminfo_scale.shmid);
    }

    xshminfo_scale.shmaddr = (char *) shmat(xshminfo_scale.shmid, 0, 0);
    if ((void *)-1 == xshminfo_scale.shmaddr) {
         p_err("X11 SHM error 4");
    }

    ximage_scale->data = xshminfo_scale.shmaddr;

    xshminfo_scale.readOnly = False;

    dbg("XShmAttach\n");
    XShmAttach(dpy, &xshminfo_scale);
    XSync(dpy, False);
    shmctl(xshminfo_scale.shmid, IPC_RMID, 0);
  }
}


static void fftv_XShmPutImage(int sx, int sy, int sxL, int syL, int dx, int dy, int dxL, int dyL)
{
#if 1
  if (sxL==dxL && syL == dyL) {
    XShmPutImage(dpy, tv_win, gc, ximage, sx, sy, dx, dy, sxL, syL, False);
  } else
#endif
  {
    int x,y;
    float scale_x = (float) sxL / dxL;
    float scale_y = (float) syL / dyL;
    int *pixPtr=(int *)ximage_scale->data;
    int sc_x_tran[SCALE_SIZE_X];

    dbg("%d %d %d %d  %d %d %d %d  %.2f %.2f\n",sx,sy, sxL, syL, dx, dy, dxL, dyL,
        scale_x, scale_y);

    for(x=0; x < dxL; x++) {
      sc_x_tran[x] = scale_x * x + sx;
    }

    int x_left = SCALE_SIZE_X - dxL;

    for(y=0; y < dyL; y++) {
      int sc_y = scale_y * y  + sy;
      int *line_start = ((int *)ximage->data) + sc_y * user_grab_sizeX;  // RGB32 format

      for(x=0; x < dxL; x++) {
        // RGB32 format
        *(pixPtr++) = line_start[sc_x_tran[x]];
      }

      pixPtr+= x_left;
    }

    XShmPutImage(dpy, tv_win, gc, ximage_scale, 0, 0, dx, dy, dxL, dyL, False);

    XFlush(dpy);
  }
}


#if 1
void put_xshm_image()
{
  if (!disp_rectN) {
      fftv_XShmPutImage(0, 0, user_grab_sizeX, user_grab_sizeY, 0, 0, win_xl, win_yl);
    return;
  }

//  dbg("disp_rectN:%d\n", disp_rectN);

  int i;
  for(i=0; i< disp_rectN; i++) {
    RECT *prect = &disp_rect[i];
    int srcx0, srcy0;
    int destsizeX = prect->x1 - prect->x0 + 1;
    int destsizeY = prect->y1 - prect->y0 + 1;
    int srcsizeX, srcsizeY;

    if (user_grab_sizeX == win_xl && user_grab_sizeY == win_yl) {
      srcsizeX = destsizeX; srcsizeY = destsizeY;
      srcx0 = prect->x0; srcy0 = prect->y0;
    } else {
      int srcx1, srcy1;
      srcx0 = (float)prect->x0 * user_grab_sizeX / win_xl;
      srcx1 = (float)prect->x1 * user_grab_sizeX / win_xl;
      srcy0 = (float)prect->y0 * user_grab_sizeY / win_yl;
      srcy1 = (float)prect->y1 * user_grab_sizeY / win_yl;
      srcsizeX = srcx1 - srcx0;
      srcsizeY = srcy1 - srcy0;

      if (srcsizeX > user_grab_sizeX)
        srcsizeX = user_grab_sizeX;
      if (srcsizeY > user_grab_sizeY)
        srcsizeY = user_grab_sizeY;
    }

    if (srcsizeY<3 || destsizeY<3)
      continue;

    fftv_XShmPutImage(srcx0, srcy0, srcsizeX, srcsizeY,
                 prect->x0, prect->y0, destsizeX, destsizeY);

  }
}
#endif

// image must be in xvimage->data
void put_xv_image()
{
#if 0
  return;
#endif
  if (!disp_rectN) {
//    dbg("put_xv_image full\n");
    XvPutImage(dpy, xvport, tv_win, gc, xvimage,
                 0, 0, user_grab_sizeX, user_grab_sizeY, /* source */
                 0, 0, win_xl, win_yl /* dest */);
    return;
  }

  int i;
  for(i=0; i< disp_rectN; i++) {
    RECT *prect = &disp_rect[i];

    int srcx0, srcy0;
    int destsizeX = prect->x1 - prect->x0 + 1;
    int destsizeY = prect->y1 - prect->y0 + 1;
    int srcsizeX, srcsizeY;

    if (user_grab_sizeX == win_xl && user_grab_sizeY == win_yl) {
      srcsizeX = destsizeX; srcsizeY = destsizeY;
      srcx0 = prect->x0; srcy0 = prect->y0;
    } else {
      int srcx1, srcy1;
      srcx0 = (float)prect->x0 * user_grab_sizeX / win_xl;
      srcx1 = (float)prect->x1 * user_grab_sizeX / win_xl;
      srcy0 = (float)prect->y0 * user_grab_sizeY / win_yl;
      srcy1 = (float)prect->y1 * user_grab_sizeY / win_yl;
      srcsizeX = srcx1 - srcx0;
      srcsizeY = srcy1 - srcy0;

      if (srcsizeX > user_grab_sizeX)
        srcsizeX = user_grab_sizeX;
      if (srcsizeY > user_grab_sizeY)
        srcsizeY = user_grab_sizeY;
    }

    XvPutImage(dpy, xvport, tv_win, gc, xvimage,
               srcx0, srcy0, srcsizeX, srcsizeY,
               prect->x0, prect->y0, destsizeX, destsizeY);
  }
}


#include <sys/mman.h>


extern int b_v4l2;
#define REQ_BUF_SZ 4
struct v4l2_buffer buf_v4l2[REQ_BUF_SZ];
static char *v4l2_image_buf[REQ_BUF_SZ];


void fftv_grab_close()
{
//  dbg("fftv_grab_close\n");

  if (b_v4l2) {
    int i;

    enum v4l2_buf_type type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    ioctl(vfd, VIDIOC_STREAMOFF, &type);

    for (i = 0; i < REQ_BUF_SZ; i++) {
      if (v4l2_image_buf[i]) {
        munmap(v4l2_image_buf[i], buf_v4l2[i].length);
        v4l2_image_buf[i] = NULL;
      }
    }

  } else {
    if (!video_buf)
      return;
    munmap(video_buf, gb_buffers.size);
    video_buf = NULL;
  }
}

static void q_buf(int i)
{
  if (ioctl(vfd, VIDIOC_QBUF, &buf_v4l2[i]) < 0)
    p_err("init_v4l2_mbuf VIDIOC_QBUF");
}

static void v4l2_mbuf_init()
{
  struct v4l2_requestbuffers rb;
  struct v4l2_format vform;

  bzero(&vform, sizeof(vform));

  vform.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

  if (fftv_grab_fmt == FFTV_I420)
    vform.fmt.pix.pixelformat = V4L2_PIX_FMT_YUV420;
  else
  if (fftv_grab_fmt == FFTV_RGB24)
    vform.fmt.pix.pixelformat = V4L2_PIX_FMT_RGB24;

  vform.fmt.pix.field = V4L2_FIELD_ANY;
  vform.fmt.pix.width = grab_sizeX;
  vform.fmt.pix.height = grab_sizeY;
  if (ioctl(vfd, VIDIOC_S_FMT, &vform) < 0) {
    perror("VIDIOC_S_FMT");
    p_err("v4l2_mbuf_init VIDIOC_S_FMT %d %d", grab_sizeX, grab_sizeY);
  }


  rb.count = REQ_BUF_SZ;
  rb.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  rb.memory = V4L2_MEMORY_MMAP;

  if (ioctl(vfd, VIDIOC_REQBUFS, &rb) != 0) {
      p_err("init_v4l2_mbuf: 1 VIDIOC_REQBUFS");
  }

  if (rb.count < 2) {
      p_err("init_v4l2_mbuf: 2  %d", rb.count);
  }

  enum v4l2_buf_type type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

  int i;
  for (i = 0; i < rb.count; i++) {
     buf_v4l2[i].index  = i;
     buf_v4l2[i].type   = V4L2_BUF_TYPE_VIDEO_CAPTURE;
     buf_v4l2[i].memory = V4L2_MEMORY_MMAP;

     if (ioctl(vfd, VIDIOC_QUERYBUF, &buf_v4l2[i]) == -1)
         p_err("init_v4l2_mbuf VIDIOC_QUERYBUF");
     v4l2_image_buf[i] = mmap(NULL, buf_v4l2[i].length,
                              PROT_READ | PROT_WRITE, MAP_SHARED,
                              vfd, buf_v4l2[i].m.offset);
     if (MAP_FAILED == v4l2_image_buf[i]) {
         p_err("init_v4l2_mbuf mmap");
     }

     q_buf(i);
  }


  if (ioctl(vfd, VIDIOC_STREAMON, &type) != 0) {
      p_err("init_v4l2_mbuf: VIDIOC_STREAMON");
  }
}



static void v4l_mbuf_init()
{
   int ret = ioctl(vfd,VIDIOCGMBUF, &gb_buffers);

   dbg("frames:%d\n", gb_buffers.frames);

   if (ret < 0) {
     p_err("VIDIOCGMBUF error %d\n",ret);
   }

   video_buf = mmap(NULL, gb_buffers.size, PROT_READ|PROT_WRITE,MAP_SHARED, vfd, 0);
   if ((guchar*)-1 == video_buf) {
       p_err("mmap failed");
   }

   struct video_picture vpict;
   bzero(&vpict,sizeof(vpict));
   ioctl(vfd,VIDIOCGPICT,&vpict);

   gb_buf.height = grab_sizeY;
   gb_buf.width = grab_sizeX;

   if (fftv_grab_fmt == FFTV_I420)
     gb_buf.format = VIDEO_PALETTE_YUV420P;
   if (fftv_grab_fmt == FFTV_RGB24)
     gb_buf.format = VIDEO_PALETTE_RGB24;

   gb_buf.frame = 0;
   dbg("init_v4l_mbuf depth:%d  pal:%d\n", vpict.depth, vpict.palette);
}

extern int border_adj_step_val;


void init_v4l_mbuf()
{
   dbg("init_v4l_mbuf..\n");
   fftv_grab_close();

   if (TV_WIN_DISP_MODE_GRAB)
     border_adj_step_val = 8;  // grab width of bt-878 must be multiple of 16
   else
     border_adj_step_val = 1;

   calc_grab_crop_win_X(user_grab_sizeX, &visible_ofs_x0,  &visible_ofs_x1);
   set_grab_frame_size();

   if (b_v4l2)
     v4l2_mbuf_init();
   else
     v4l_mbuf_init();
}


#define DBG (0)

// YUV420P cropping;
static void frame_crop_I420(char *image)
{
  int ofs_x0 = visible_ofs_x0;
  int imageN=0;
  char *out_image = (tv_win_refresh_mode == TV_WIN_REFRESH_GRAB) ? xvimage->data : user_grab_image;

  // ofs_x0, ofs_x1, can only be even
  int y;

  char *ptr = image + ofs_x0;

  for(y = 0; y <  user_grab_sizeY; y++) {
    memcpy(out_image + imageN , ptr, user_grab_sizeX);

    ptr+= grab_sizeX;
    imageN += user_grab_sizeX;
  }

  // copy cb
  ptr = image + grab_pixN + ofs_x0/2;

  for(y = 0; y <  user_grab_sizeY/2; y++) {
    memcpy(out_image + imageN , ptr, user_grab_sizeX/2);

    ptr+= grab_sizeX / 2;
    imageN += user_grab_sizeX/2;
  }

#if DBG
  if (imageN != pixN + user_grab_pixN/4)
    p_err("error 1 %d", imageN);
#endif

  // copy cr
  ptr = image + grab_pixN + grab_pixN/4 + ofs_x0/2;

  for(y = 0; y <  user_grab_sizeY/2; y++) {
    memcpy(out_image + imageN , ptr, user_grab_sizeX/2);

    ptr+= grab_sizeX / 2;
    imageN += user_grab_sizeX/2;
  }

//  dbg("imageN:%d user_grab_imageN:%d\n", imageN,  user_grab_imageN);
#if DBG
  if (imageN != user_grab_pixN*3/2)
    p_err("error found");
#endif
}



// RGB24 cropping, also coverted to RGB32
static void frame_crop_RGB24(char *image, int b_out24)
{
  int ofs_x0 = visible_ofs_x0;
  char *out_image = ximage ? ximage->data : NULL;
  char *out_image24 = user_grab_image;
  int y;
  int rowbytes = grab_sizeX*PIX_BYTES_24;
  char *st_ptr = image + ofs_x0*PIX_BYTES_24;

  for(y = 0; y <  user_grab_sizeY; y++) {
    char *ptr = st_ptr;
    int x;

    if (ximage)
      for(x=0; x < user_grab_sizeX; x++) {
        *((int *)out_image) = *((int *)ptr);
        out_image+=4;
        ptr+=3;
      }

    if (b_out24) {
      memcpy(out_image24, st_ptr, PIX_BYTES_24 * user_grab_sizeX);
      out_image24+=PIX_BYTES_24 * user_grab_sizeX;
    }

    st_ptr += rowbytes;
  }
}



static gchar *fftv_read_frame_v4l2(int b_out24)
{
  gchar *imageptr;

#if 1
  struct pollfd pofd;

  pofd.fd = vfd;
  pofd.events = POLLIN;
  pofd.revents = 0;

  if (poll(&pofd, 1, 1000) <= 0)
    p_err("fftv_read_frame 1, cannot poll vfd");
#endif

  struct v4l2_buffer buf2;

  bzero(&buf2, sizeof(buf2));
  buf2.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  buf2.field = V4L2_FIELD_ANY;
  buf2.flags = 0;

  if (ioctl(vfd, VIDIOC_DQBUF, &buf2) == -1) {
    perror("VIDIOC_DQBUF ");
    p_err("fftv_read_frame 2");
  }

  imageptr = v4l2_image_buf[buf2.index];

  if ((visible_ofs_x0 || visible_ofs_x1)) {
    if (fftv_grab_fmt == FFTV_I420)
      frame_crop_I420(imageptr);
    else
      frame_crop_RGB24(imageptr, b_out24);

    if (tv_win_refresh_mode == TV_WIN_REFRESH_GRAB) {
      if (fftv_grab_fmt == FFTV_I420)
        imageptr = xvimage->data;
      else
        imageptr = user_grab_image;
    } else
        imageptr = user_grab_image;
  } else {
    if (tv_win_refresh_mode == TV_WIN_REFRESH_GRAB) {
      if (fftv_grab_fmt == FFTV_I420) {
        memcpy(xvimage->data, imageptr, grab_bytesN);
        imageptr = xvimage->data;
      } else
      if (fftv_grab_fmt == FFTV_RGB24) {
        frame_crop_RGB24(imageptr, b_out24);
        imageptr = ximage->data;
      }
      else {
        memcpy(ximage->data, imageptr, grab_bytesN);
        imageptr = ximage->data;
      }
    } else {
      memcpy(user_grab_image, imageptr, grab_bytesN);
      imageptr =  user_grab_image;
    }
  }

  q_buf(buf2.index);
#if 0
  dbg("read %d %x\n", buf2.index, buf2.flags);
#endif
  return imageptr;
}


gchar *fftv_read_frame(int b_out24)
{
  gchar *imageptr;

  if (b_v4l2) {
    return fftv_read_frame_v4l2(b_out24);
  }

  gb_buf.frame = (gb_frame + 1) % gb_buffers.frames;

  if (ioctl(vfd, VIDIOCMCAPTURE, &gb_buf) < 0) {
      if (errno == EAGAIN)
          fprintf(stderr,"Cannot Sync\n");
      else
          perror("fftv_read_frame:VIDIOCMCAPTURE");

      return NULL;
  }

  while (ioctl(vfd, VIDIOCSYNC, &gb_frame) < 0 &&
         (errno == EAGAIN || errno == EINTR));

  imageptr = video_buf + gb_buffers.offsets[gb_frame];
  gb_frame = gb_buf.frame;


  if ((visible_ofs_x0 || visible_ofs_x1)) {
    if (fftv_grab_fmt == FFTV_I420)
      frame_crop_I420(imageptr);
    else
      frame_crop_RGB24(imageptr, b_out24);

    if (tv_win_refresh_mode == TV_WIN_REFRESH_GRAB) {
      if (fftv_grab_fmt == FFTV_I420)
        return xvimage->data;
      else
        return user_grab_image;
    } else
      return user_grab_image;
  } else {
    if (tv_win_refresh_mode == TV_WIN_REFRESH_GRAB) {
      if (fftv_grab_fmt == FFTV_I420) {
        memcpy(xvimage->data, imageptr, grab_bytesN);
        return xvimage->data;
      } else
      if (fftv_grab_fmt == FFTV_RGB24) {
        frame_crop_RGB24(imageptr, b_out24);
        return ximage->data;
      }
      else {
        memcpy(ximage->data, imageptr, grab_bytesN);
        return ximage->data;
      }
    } else {
      memcpy(user_grab_image, imageptr, grab_bytesN);
      return user_grab_image;
    }
  }
}



void refresh_grab_tv_win()
{
#if 1
  if (!tv_win)
    return;

  guchar *imageptr = fftv_read_frame(FALSE);

  if (fftv_grab_fmt == FFTV_I420)
    put_xv_image(imageptr);
  else
    put_xshm_image(imageptr);

#endif
  return;
}

static int local_win_x, local_win_y, local_win_sizeX, local_win_sizeY;

void grab_draw_local_win(int x, int y, int sizeX, int sizeY)
{
  fftv_read_frame(FALSE);

  local_win_x = x;  local_win_y = y;
  local_win_sizeX = sizeX;   local_win_sizeY = sizeY;

  if (fftv_grab_fmt == FFTV_I420)
    XvPutImage(dpy, xvport, tv_win, gc, xvimage,
                   0, 0, grab_sizeX, grab_sizeY, /* source */
                   x, y, sizeX, sizeY /* dest */);
  else {
    fftv_XShmPutImage(0, 0, grab_sizeX, grab_sizeY, /* source */
                   x, y, sizeX, sizeY /* dest */);
  }
}


gboolean refresh_local_win()
{
  if (!tv_win)
    return TRUE;
#if 0
  dbg("refresh_local_win: %d %d %d %d\n",
    local_win_x, local_win_y, local_win_sizeX, local_win_sizeY);
#endif
  grab_draw_local_win(local_win_x, local_win_y, local_win_sizeX, local_win_sizeY);
  return TRUE;
}


void init_Xv_port()
{
  XvAdaptorInfo *adaptorArr, *padaptor;
  int adaptorArrN, i, j, k;

#if 0
  goto no_xv;
#endif

  if (XvQueryAdaptors(dpy, root, &adaptorArrN,&adaptorArr) != Success) {
    dbg("cannot XvQueryAdaptors error\n");
    goto no_xv;
  }

  for (i=0; i<adaptorArrN; i++) {
    padaptor = &adaptorArr[i];

    if (!((padaptor->type & XvInputMask) && (padaptor->type & XvImageMask)) )
      continue;

    dbg("padaptor->num_ports: %d\n", padaptor->num_ports);

    for (j=0; j<padaptor->num_ports;j++) {
      XvImageFormatValues *imgFormatArr=NULL;
      int imgFormatArrN=0;

      xvport = padaptor->base_id + j;
      imgFormatArr = XvListImageFormats(dpy, xvport, &imgFormatArrN);

      dbg("imgFormatArrN:%d\n", imgFormatArrN);

      if (!imgFormatArr || !imgFormatArrN)
        continue;

      if (XvGrabPort(dpy, xvport, CurrentTime) != Success)
        continue;

      for (k=0; k< imgFormatArrN; k++) {
        int vid = imgFormatArr[k].id;
        char *tt = (char *)&vid;

        dbg("zzzzzzzzzzzz %x %c%c%c%c\n",vid, tt[0],tt[1],tt[2],tt[3]);
      }
#if 1
      for (k=0; k< imgFormatArrN; k++) {
        xvmode = imgFormatArr[k].id;
        if (fftv_grab_fmt == FFTV_I420 && !memcmp(&xvmode,"I420",4)) {
          goto xvport_found;
        }
      }
#endif
      XvUngrabPort(dpy, xvport, CurrentTime);
    }
  }

  XvFreeAdaptorInfo(adaptorArr);
no_xv:
  dbg("YUV420P(I420) format is not found in your XVideo interface, RGB24/MIT_SHM will be used");
  fftv_grab_fmt = FFTV_RGB24;
//  tv_win_disp_mode = TV_GRAB_MIT_SHM;
  return;

xvport_found:
  XvFreeAdaptorInfo(adaptorArr);
  dbg("YUV420P(I420) format xvport:%x\n", xvport);
}
