#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>

#include <linux/videodev.h>
#include <unistd.h>
#include <sys/time.h>
#include <time.h>
#include <setjmp.h>

#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>
#include <X11/Xproto.h>
#include <X11/extensions/XTest.h>
#include <X11/keysym.h>
#include <X11/extensions/xf86vmode.h>

#include "reminder.h"

#include <string.h>
#include <ctype.h>

#include "tv-freq.h"
#include "fftv.h"
#include "xkey.h"
#include "vidstr.h"
#include "tv-win.h"

static void set_dpy_visibily_clips();
void disp_channel_name_idx();
static gboolean cb_unmap();
static gboolean key_eve(GtkWidget *widget, GdkEventKey *event, gpointer user_data);
static gboolean VisibilityNotify_event(GtkWidget *widget, GdkEvent *event, gpointer user_data);
static gboolean cb_ConfEve(GtkWidget *widget,GdkEventConfigure *event,
                        gpointer user_data);
#if LIRC
extern int lirc_fd;
#endif
extern int rem_sockfd;
Display *dpy;
static GdkWindow *gdk_root;
static GdkCursor* xcursor;
Window frame_win;
void fullscreen_toggle();


#if 0
TV_WIN_DISP_MODE tv_win_disp_mode = (TV_OVERLAY | TV_GRAB_XV_RECORD);
#elif 0
TV_WIN_DISP_MODE tv_win_disp_mode = (TV_GRAB_XV | TV_GRAB_XV_RECORD);
#else
TV_WIN_DISP_MODE tv_win_disp_mode = (TV_OVERLAY | TV_OVERLAY_RECORD);
#endif

TV_WIN_REFRESH_MODE tv_win_refresh_mode;
static gboolean b_change_tv_win_refresh_mode = FALSE;

void set_tv_win_refresh_mode(TV_WIN_REFRESH_MODE mode)
{
  if (tv_win_refresh_mode == mode)
    return;

  b_change_tv_win_refresh_mode = TRUE;
  tv_win_refresh_mode = mode;
}


static int xerror_handler(Display *d, XErrorEvent *eve)
{
  return 0;
}

void p_err(char *fmt,...)
{
	va_list args;

	va_start(args, fmt);
	fprintf(stderr,"fftv:");
	vfprintf(stderr, fmt, args);
	va_end(args);
	fprintf(stderr,"\n");
	exit(-1);
}


void error(char *fmt,...)
{
	va_list args;

	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	va_end(args);
}


void dbg(char *fmt,...)
{
	va_list args;

	va_start(args, fmt);
	vprintf(fmt, args);
	fflush(stdout);
	va_end(args);
}


long current_time()
{
  struct timeval timev;
  struct timezone tz;

  gettimeofday(&timev,&tz);
  return timev.tv_sec;
}


void save_original_dpy_mode();
void init_xk_ff();


void init_X_0();
void load_disp_opt();
void mixerInit();

void init_X_arg(int *argc, char ***argv)
{
   if (!getenv("DISPLAY")) {

     putenv("DISPLAY=:0.0");
#if 0
     sprintf(tt,"XAUTHORITY=%s/.Xauthority", getenv("HOME"));
     putenv(tt);
#endif
   }

#if LIRC
   open_lirc();
#endif
   init_rem_serv();
   mixerInit();

   if (fftv_mode & FFTV_MODE_RADIO)
     set_fftv_cmd_state("FM-radio");

   init_xk_ff();

//   dbg("init_X_arg rem_sockfd: %d\n",rem_sockfd);
   load_disp_opt();

   gtk_init (argc, argv);

   init_X_0();

   save_original_dpy_mode();
}


static Colormap colormap;
long foreground, background;
long color_red, color_pink;
long white, grey, black, cyan, blue, brown, gold;
static char *fg_color="yellow",*bg_color="blue";
Window tv_win;
GtkWidget *gtv_win;
GdkWindow *gdk_tv_win;
Window root;
GC gc, gc_sfont, gc_rec;
GdkGC *g_gc;

static int screen_no;
static int xsockfd;
int win_borderless, owin_borderless;
static int b_pointer_timer;
gboolean b_fftv_invisible;
int full_win;
int full_win640;
char *geomstr="640x480";
int win_xl=640, win_yl=480,win_x=100,win_y=100;
static int owin_xl=640,owin_yl=480;
static int owin_x=100,owin_y=100;
int timeout_tag_refresh_grab_tv_win;
int border_adj_step_val=1;


void disp_border_value();

void tv_video_shrink_left()
{
  if (pchan_current->ofs_x0 < 100)
    pchan_current->ofs_x0+=border_adj_step_val;
  save_channel(pchan_current);

  disp_border_value();
  if (TV_WIN_DISP_MODE_GRAB)
    init_v4l_mbuf();
}

void tv_video_shrink_right()
{
  if (pchan_current->ofs_x1 < 100)
    pchan_current->ofs_x1+=border_adj_step_val;
  save_channel(pchan_current);

  disp_border_value();
  if (TV_WIN_DISP_MODE_GRAB)
    init_v4l_mbuf();
}

void tv_video_grow_left()
{
  pchan_current->ofs_x0-=border_adj_step_val;
  if (pchan_current->ofs_x0 < 0)
    pchan_current->ofs_x0 = 0;

  save_channel(pchan_current);
  set_vwin();

  disp_border_value();
  if (TV_WIN_DISP_MODE_GRAB)
    init_v4l_mbuf();
}

void tv_video_grow_right()
{
  pchan_current->ofs_x1-=border_adj_step_val;
  if (pchan_current->ofs_x1 < 0)
    pchan_current->ofs_x1 = 0;

  save_channel(pchan_current);

  disp_border_value();

  if (TV_WIN_DISP_MODE_GRAB)
    init_v4l_mbuf();
}


void tv_video_shrink_bottom()
{
  if (pchan_current->ofs_y1 < 100)
    pchan_current->ofs_y1++;
  save_channel(pchan_current);

  disp_border_value();
}

void tv_video_shrink_top()
{
  if (pchan_current->ofs_y0 < 100)
    pchan_current->ofs_y0++;
  dbg("pchan_current->ofs_y0:%d\n", pchan_current->ofs_y0);
  save_channel(pchan_current);

  disp_border_value();
}

void tv_video_grow_bottom()
{
  if (pchan_current->ofs_y1 > 0)
    pchan_current->ofs_y1--;

  save_channel(pchan_current);
  disp_border_value();
}

void tv_video_grow_top()
{
  if (pchan_current->ofs_y0 > 0)
    pchan_current->ofs_y0--;

  save_channel(pchan_current);
  disp_border_value();
}



struct {
  char *color_name;
  long *color_addr;
} color_arr[] = {
  {"white",&white},
  {"black",&black},
  {"grey",&grey},
  {"red",&color_red},
  {"pink",&color_pink},
  {"cyan",&cyan},
  {"blue",&blue},
  {"brown",&brown},
  {"gold",&gold},
};

int color_arrN=sizeof(color_arr)/sizeof(color_arr[0]);

int pic_adj_menu_mode;

static GdkPixmap *bm_off;
static GdkCursor *ptr_off;

static void free_mouse_pointer()
{
  if (bm_off) {
    gdk_cursor_unref(ptr_off);
    gdk_pixmap_unref(bm_off);
    bm_off=0;
  }
}

static void move_mouse_xy(int x, int y)
{
  gdk_flush();
  dbg("move mouse %d %d  tv_win:%x\n",x,y,tv_win);
  if (!tv_win)
    return;
  XWarpPointer(dpy,None,tv_win,0,0,0,0,x,y);
}

static void init_hide_mouse_pointer()
{
  static unsigned char bm_null[] = { 0,0,0,0, 0,0,0,0 };
  GdkColor gblack = {0, 0, 0, 0};

  free_mouse_pointer();

  bm_off = gdk_bitmap_create_from_data(NULL,bm_null, 8,8);
  ptr_off = gdk_cursor_new_from_pixmap(bm_off, bm_off,&gblack, &gblack,0, 0);
}

static void hide_mouse_pointer()
{
//  dbg("a hide mouse ptr\n");
  if (!gdk_tv_win)
    return;
  b_pointer_timer=0;
  gdk_window_set_cursor(gdk_tv_win, ptr_off);
//  dbg("b hide mouse ptr\n");
}

XFontStruct *afont, *afont_small;

static void create_gc()
{
  XGCValues gcv, gcv_sfont, gcv_rec;
  gcv.foreground = foreground;
  gcv.background = background;
  gcv.graphics_exposures = False;
  gcv.font=afont->fid;
  if (gc) XFreeGC(dpy,gc);
  gc = XCreateGC(dpy,tv_win,GCForeground|GCBackground|GCFont|
  GCGraphicsExposures,&gcv);

  gcv_sfont.foreground = foreground;
  gcv_sfont.background = background;
  gcv_sfont.graphics_exposures = False;
  gcv_sfont.font=afont_small->fid;
  if (gc_sfont) XFreeGC(dpy,gc_sfont);
  gc_sfont = XCreateGC(dpy,tv_win,GCForeground|GCBackground|GCFont|
  GCGraphicsExposures,&gcv_sfont);

  gcv_rec.foreground = foreground;
  gcv_rec.background = background;
  gcv_rec.graphics_exposures = False;
  gcv_rec.font=afont_small->fid;
  if (gc_rec) XFreeGC(dpy,gc_rec);
  gc_rec = XCreateGC(dpy,tv_win,GCForeground|GCBackground|GCFont|
  GCGraphicsExposures,&gcv_rec);
}

void change_win_name(char *str)
{
  gtk_window_set_title (GTK_WINDOW(gtv_win), str);
}

extern CHANLIST *pchan_current;
static void getGlobalWinRect(Window win, int *x0, int *y0, int *x1, int *y1);
#if LIRC
void lirc_callback(gpointer data, gint source, GdkInputCondition condition);
#endif
void reminder_callback(gpointer data, gint source, GdkInputCondition condition);
void feed_xkey(int skey, int state, int chr);

int tv_win_mouse_x, tv_win_mouse_y;

static void mouse_button_callback( GtkWidget *widget,GdkEventButton *event, gpointer data)
{
  guint xkey=0;

  if (!tv_win)
    return;

//  dbg("------ mouse_button_callback\n");
  tv_win_mouse_x = event->x;
  tv_win_mouse_y = event->y;

  switch (event->button) {
    case 1:
      xkey=MOUSE_BUTTON1;
      break;
    case 2:
      xkey=MOUSE_BUTTON2;
      break;
    case 3:
      xkey=MOUSE_BUTTON3;
      break;
  }

  feed_xkey(xkey,event->state,0);
}

void mouse_scroll_callback(GtkWidget *widget,GdkEvent *event, gpointer data)
{
  GdkEventScroll *pscroll=&event->scroll;
  guint xkey=0;

  if (pscroll->direction==GDK_SCROLL_UP) {
    xkey = MOUSE_BUTTON4;
  }
  else
  if (pscroll->direction==GDK_SCROLL_DOWN) {
    xkey = MOUSE_BUTTON5;
  }

  feed_xkey(xkey,pscroll->state,0);
}


gint mouse_pointer_hide_callback(gpointer data)
{
//  dbg("---- mouse_pointer_hide_callback\n");
  b_pointer_timer=0;

  if (!(fftv_mode & FFTV_MODE_TV))
    return FALSE;

//  dbg("motion notify cb ..\n");

  hide_mouse_pointer();
  return FALSE;
}


gboolean motion_notify_callback(GtkWidget *widget, GdkEventMotion *event,
                                gpointer user_data)
{
//  dbg("------ motion_notify_callback\n");
  if (!(fftv_mode & FFTV_MODE_TV))
    return TRUE;

  if (!b_pointer_timer) {
//    dbg("motion notify ..\n");
    b_pointer_timer=1;
    gdk_window_set_cursor(gdk_tv_win,xcursor);
    gtk_timeout_add(3000,mouse_pointer_hide_callback,NULL);
  }

  return TRUE;
}

void reset_sreensaver(char *caller);

gint reset_screensaver_callback(gpointer data)
{
  if ((win_borderless && (fftv_mode & FFTV_MODE_TV))
       || record_status==RECORD_ON)
    reset_sreensaver("reset_screensaver_callback");

  return TRUE;
}

void ExposeEve0();
static void tv_win_delayed_expose();

GdkFilterReturn root_event_callback(GdkXEvent *gdkxevent,
                                    GdkEvent *event,gpointer data)
{
  XEvent *xeve=gdkxevent;
  if (b_fftv_invisible || !tv_win)
    return TRUE;

  switch (xeve->type) {
#if 1
    case VisibilityNotify:
      dbg("root visibility\n");
      break;
#endif
    case ConfigureNotify:
#if 0
      dbg("root configure\n");
#endif
      tv_win_delayed_expose();
      break;
  }

  return TRUE;
}

void init_xvisual();
gint soundcard_vol_fetch_update();


void init_X_0()
{
  dpy = GDK_DISPLAY();
  xsockfd=ConnectionNumber(dpy);
  gdk_root=gdk_get_default_root_window();
  xcursor = gdk_cursor_new(GDK_STAR);
  root=DefaultRootWindow(dpy);
  screen_no = DefaultScreen(dpy);
  init_hide_mouse_pointer();
#if LIRC
  if (lirc_fd > 0) {
     dbg("LIRC connected lircfd: %d\n", lirc_fd);
     gdk_input_add(lirc_fd,GDK_INPUT_READ, lirc_callback, NULL);
  }
  else
     dbg("!!! Warning! fftv could not connect to lircd. LIRC won't work.\n");
#endif
  gdk_input_add(rem_sockfd, GDK_INPUT_READ, reminder_callback, NULL);
  gtk_timeout_add(1000, soundcard_vol_fetch_update, NULL);

  gtk_timeout_add(60000, reset_screensaver_callback, NULL);
  gdk_window_set_events(gdk_root, GDK_VISIBILITY_NOTIFY_MASK | GDK_EXPOSURE_MASK |
                     GDK_SUBSTRUCTURE_MASK | GDK_STRUCTURE_MASK);
  gdk_window_add_filter(gdk_root,root_event_callback,NULL);
  init_xvisual();
}

void refresh_grab_tv_win();


void getTVWinXY(int *gx, int *gy);
static Window get_frame_win();

static time_t last_tv_preview_time;


gboolean cb_ExposeEve(GtkWidget *widget, GdkEventExpose *event, gpointer user_data);
void create_button_widget();
void do_exit_fftv();
void destroy_border_adj();
extern gboolean b_current_sticky;
extern GdkWindowTypeHint current_GDK_WINDOW_TYPE_HINT;

void create_tv_win(gboolean b_geomstr, gboolean b_desktop)
{
  dbg("create_tv_win\n");

  if (gtv_win) {
    destroy_border_adj();

    g_signal_handlers_block_by_func(G_OBJECT(gtv_win),G_CALLBACK(do_exit_fftv),NULL);
    gtk_widget_destroy(gtv_win);
    gtv_win=NULL;
    tv_win=(Window)NULL;
    gdk_tv_win=NULL;
    last_tv_preview_time = 0;
  }

  gtv_win = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_icon_from_file(GTK_WINDOW(gtv_win), ICON_BASENAME "/fftv.png", NULL);
  gtk_widget_set_double_buffered (gtv_win, FALSE);
  gtk_widget_set_app_paintable (gtv_win, TRUE);

  gtk_window_set_type_hint(GTK_WINDOW(gtv_win), current_GDK_WINDOW_TYPE_HINT);
  if (b_current_sticky)
    gtk_window_stick(GTK_WINDOW(gtv_win));

  gtk_widget_set_events(gtv_win,GDK_POINTER_MOTION_MASK|GDK_EXPOSURE_MASK|
    GDK_SCROLL_MASK|GDK_VISIBILITY_NOTIFY_MASK|GDK_SCROLL_MASK);

  g_signal_connect (G_OBJECT (gtv_win), "delete_event",
                      G_CALLBACK (do_exit_fftv), NULL);

  if (b_desktop) {
    extern gboolean b_current_decorated;

    win_borderless = !b_current_decorated;
  }

  if (win_borderless) {
    gtk_window_set_decorated(GTK_WINDOW(gtv_win),FALSE);
  } else {
    gtk_window_set_decorated(GTK_WINDOW(gtv_win), TRUE);
  }

  gtk_window_set_title (GTK_WINDOW(gtv_win), "fftv");

  if (b_geomstr) {
    gtk_window_parse_geometry(GTK_WINDOW(gtv_win), geomstr);
  }
  else {
    gtk_window_set_default_size(GTK_WINDOW(gtv_win), win_xl, win_yl);
    gtk_window_move(GTK_WINDOW(gtv_win),win_x,win_y);  // need  this for gnome 1.4 sawfish
  }

  gtk_widget_show_now (gtv_win);

  gdk_tv_win = gtv_win->window;
  extern gboolean b_override_redirect;
  if (b_override_redirect) {
    gdk_window_set_override_redirect(gdk_tv_win, TRUE);
    dbg("... set_override_redirect\n");
  }


  tv_win = GDK_WINDOW_XWINDOW(gdk_tv_win);

  create_gc();

  hide_mouse_pointer();

  if (pchan_current)
    change_win_name(pchan_current->name);

  g_signal_connect (G_OBJECT(gtv_win), "scroll-event",
                      G_CALLBACK (mouse_scroll_callback), NULL);

  getTVWinXY(&win_x,&win_y);
  gtk_window_get_size(GTK_WINDOW(gtv_win),&win_xl,&win_yl);

  frame_win = get_frame_win();

  dbg("---- create_tv %d %d   %d %d  frame:%x\n",win_x,win_y,win_xl,win_yl,frame_win);
  create_button_widget();


  g_signal_connect(G_OBJECT(gtv_win),"destroy", G_CALLBACK(do_exit_fftv), NULL);
  g_signal_connect(G_OBJECT(gtv_win),"configure-event", G_CALLBACK(cb_ConfEve), NULL);
  g_signal_connect(G_OBJECT(gtv_win),"unmap-event", G_CALLBACK(cb_unmap), NULL);
  g_signal_connect(G_OBJECT(gtv_win),"key-press-event", G_CALLBACK(key_eve), NULL);
  g_signal_connect(G_OBJECT(gtv_win),"button-press-event", G_CALLBACK(mouse_button_callback), NULL);
  g_signal_connect(G_OBJECT(gtv_win),"visibility-notify-event",
     G_CALLBACK(VisibilityNotify_event), NULL);
#if 1
  g_signal_connect(G_OBJECT(gtv_win),"expose-event", G_CALLBACK(cb_ExposeEve), NULL);
#endif
  g_signal_connect(G_OBJECT(gtv_win),"motion-notify-event",
     G_CALLBACK(motion_notify_callback), NULL);

}


static void create_colors()
{
  XColor color;
  int i;

  if (!XParseColor(dpy,colormap,fg_color, &color))
    p_err("invalid color %s",fg_color);
  if (!XAllocColor(dpy,colormap,&color))
    p_err("can't allocate color %s, will use private colormap.", fg_color);
  foreground=color.pixel;

  if (!XParseColor(dpy,colormap,bg_color, &color))
    p_err("invalid color %s",bg_color);
  if (!XAllocColor(dpy,colormap,&color))
    p_err("can't allocate color %s, will use private colormap.", bg_color);
  background=color.pixel;

  for(i=0;i<color_arrN;i++) {
    if (!XParseColor(dpy,colormap,color_arr[i].color_name, &color))
      p_err("invalid color %s",color_arr[i].color_name);
    if (!XAllocColor(dpy,colormap,&color))
      p_err("can't allocate color %s ", color_arr[i].color_name);
    *color_arr[i].color_addr=color.pixel;
  }

}

void show_tv_win()
{
  if (!gtv_win)
    create_tv_win(TRUE, FALSE);
  gtk_widget_show(gtv_win);
}

#if 0
static char *FontName="-adobe-courier-bold-r-normal--40-0-0-0-m-0-iso8859-9";
#else
// Thanks to Job Dio'genes Ribeiro Borges
static char *FontName="-*-courier-bold-r-normal--40-0-0-0-m-0-*-*";
#endif
static char *sFontName="12x24";
int fwidth,fheight;
int sfwidth,sfheight;


static void load_font()
{

  if (afont) XFreeFont(dpy, afont);
  if (!(afont=XLoadQueryFont(dpy, FontName)))
    p_err("Cannot load ascii font %s", FontName);
  fwidth=afont->max_bounds.width;
  fheight=afont->ascent + afont->descent;

  if (afont_small) XFreeFont(dpy, afont_small);
  if (!(afont_small=XLoadQueryFont(dpy, sFontName)))
    p_err("Cannot load ascii font %s", sFontName);
  sfwidth=afont_small->max_bounds.width;
  sfheight=afont_small->ascent + afont_small->descent;

}

void show_radio_win();

void init_X_win(int argc, char **argv)
{
  owin_borderless=win_borderless;
  colormap = DefaultColormap(dpy,screen_no);

  load_font();

  create_colors();

  if (fftv_mode & FFTV_MODE_TV)
    create_tv_win(TRUE, FALSE);
  else
    show_radio_win();
}


void getTVWinXY(int *gx, int *gy)
{
  gdk_window_get_origin(gdk_tv_win,&win_x,&win_y);
}


static void getGlobalWinRect(Window win, int *x0, int *y0, int *x1, int *y1)
{
  Window ow;
  XWindowAttributes watt;

  XGetWindowAttributes(dpy,win,&watt);

  XTranslateCoordinates(dpy,win,root,0,0,x0,y0,&ow);

  *x1=*x0 + watt.width;
  *y1=*y0 + watt.height;
}

void clear_tv_win_move_timeout();
void refreshTree(Window inwin, int tvx0, int tvy0, int tvx1,int tvy1, int depth, gboolean my_own);

void fftv_invisible()
{
  void *olderr = XSetErrorHandler(xerror_handler);

  tv_pic_off();
  clear_tv_win_move_timeout();

  refreshTree(root, win_x, win_y, win_x+win_xl, win_y+win_yl, 0, False);

  b_fftv_invisible=1;

  XSetErrorHandler(olderr);
}


static void set_dpy_visibily_clips()
{
  Window r_root,r_parent;
  Window *child;
  int childn;
  int tvx0,tvy0,tvx1,tvy1,i;

  void *olderr = XSetErrorHandler(xerror_handler);

  if (!tv_win)
    return;

  gdk_flush();

  clear_clips();

//  dbg("--- set_dpy_visibily_clips\n");
  frame_win = get_frame_win();
  getGlobalWinRect(tv_win,&tvx0,&tvy0,&tvx1,&tvy1);
  win_x=tvx0; win_y=tvy0;
//  dbg("tv: %d %d %d %d\n",tvx0,tvy0,tvx1,tvy1);


  XQueryTree(dpy,root,&r_root,&r_parent,&child,&childn);

  for(i=0;i<childn && child[i]!=frame_win;i++);

  for(i++;i<childn;i++) {
    Window win=child[i];
    int x0,y0,x1,y1;
    XWindowAttributes watt;
    int cx0,cy0,cx1,cy1,cxl,cyl;
    int cx,cy;
    struct video_clip *pclip;

    XGetWindowAttributes(dpy, win, &watt);

    if (!(watt.map_state & IsViewable))
      continue;
    getGlobalWinRect(win,&x0,&y0,&x1,&y1);

    if (x0>tvx1 || y0>tvy1 || x1<tvx0 || y1<tvy0)
      continue;

    cx0=MAX(x0,tvx0);
    cy0=MAX(y0,tvy0);
    cx1=MIN(x1,tvx1);
    cy1=MIN(y1,tvy1);

    cxl=cx1 - cx0;
    if (cxl > max_vxl)
      cxl=max_vxl;
    cyl=cy1 - cy0;
    if (cyl > max_vyl)
      cyl=max_vyl;

    cx=cx0 - tvx0;
    cy=cy0 - tvy0;

    if (cx>=max_vxl || cy>=max_vyl)
      continue;

    pclip=incr_clips();
    pclip->x=cx; pclip->y=cy;
    pclip->width=cxl; pclip->height=cyl;
    pclip->next=NULL;
#if 0
    dbg("-- x0:%d x1:%d   tx0:%d tx1:%d\n",x0,x1,tvx0,tvx1);
    dbg("clip  %d %d %d %d\n",pclip->x, pclip->y,pclip->width,pclip->height);
#endif
  }


  XFree(child);

  XSetErrorHandler(olderr);
}

static char remkeys[4];
static int remkeysN;
int record_status=0;
void setChannelFreq_rno(int rno);
void hide_tv_list_window(),hide_tv_ui_window();


static void remote_control_channel(char *remkeys, int *remkeysN)
{
  int rno;

  rno=atoi(remkeys);
  *remkeysN=0;
  tv_preview_mode=0;
  setChannelFreq_rno(rno);
  if (fftv_mode & FFTV_MODE_TV) {
    set_fftv_cmd_state("Normal");
    if (full_win)
      fullscreen_toggle();
  }
  else
  if (fftv_mode & FFTV_MODE_RADIO)
    set_fftv_cmd_state("FM-radio");
}


#define K100 '.'

#define REM_CLEAR_SEC (5)
static int feed_digit(char achar)
{
   char stname[32];
   int stnameN;

   if (remkeysN==0 && achar==K100) {
     remkeys[remkeysN++]=achar;
     strcpy(stname,"1--");
     stnameN=3;
     stname[stnameN]=0;
     draw_channel_name(stname,REM_CLEAR_SEC);
     return 1;
   } else
   if (achar>='0' && achar<='9') {
     remkeys[remkeysN++]=achar;
     remkeys[remkeysN]=0;
     if (remkeysN==3 && remkeys[0]==K100) {
       remkeys[0]='1';
       remote_control_channel(remkeys,&remkeysN);
     } else
     if (remkeys[0]!=K100) {
       if (remkeysN==2) {
         remote_control_channel(remkeys,&remkeysN);
       } else {
         stname[0]=remkeys[0];
         stname[1]='-';
         stnameN=2;
         tv_preview_mode=0;
         stname[stnameN]=0;
         draw_channel_name(stname,REM_CLEAR_SEC);
       }
     } else {
       if (remkeysN==2) {
         stname[0]='1';
         stname[1]=remkeys[1];
         stname[2]='-';
         stnameN=3;
         stname[stnameN]=0;
         tv_preview_mode=0;
         draw_channel_name(stname,REM_CLEAR_SEC);
       }
     }

     return 1;
   }

   return 0;
}

void feed_digit_int(int val)
{
  if (val<10)
    feed_digit('0'+val);
  else
  if (val==100)
    feed_digit(K100);
}


void clear_remkeys()
{
  pch_str->xl=0;
  remkeysN=0;
}

extern gboolean b_ch_monitor;
void save_current_channel_name();
void disp_rec_status();
void disp_rec_output_status();
static void refresh_covering_window();


static gboolean key_eve(GtkWidget *widget, GdkEventKey *event, gpointer user_data)
{
    gchar chr=0;
//    dbg("--- key_eve\n");
    if (event->string)
      chr=event->string[0];

    feed_xkey(event->keyval,event->state,chr);

    return TRUE;
}


#define SKIP_PERIOD (3)

void draw_tv_preview();
void update_pre_win_size();


static int tag_tv_win_move_timeout;
static gboolean tv_win_move_timeout()
{
  tag_tv_win_move_timeout=0;

  set_vwin();

  return FALSE;
}

void clear_tv_win_move_timeout()
{
    if (tag_tv_win_move_timeout) {
      gtk_timeout_remove(tag_tv_win_move_timeout);
      tag_tv_win_move_timeout = 0;
    }
}

void add_tv_preview_timer(int msec);

static gboolean cb_ConfEve(GtkWidget *widget,GdkEventConfigure *event,
                     gpointer user_data)
{
#if 0
  dbg("--- ConfEve %x %x\n",gtv_win,tv_win);
#endif
  tv_pic_off();

  if (b_fftv_invisible)
    return FALSE;

  if (!tv_win)
    return FALSE;

  set_dpy_visibily_clips();

  gtk_window_get_size(GTK_WINDOW(gtv_win),&win_xl,&win_yl);
  getTVWinXY(&win_x,&win_y);

#if 0
  dbg("Configure xl:%d yl:%d  x:%d y:%d\n",win_xl,win_yl,win_x,win_y);
#endif
  if (tv_preview_mode) {
    update_pre_win_size();

    add_tv_preview_timer(500);
  }
  else {
    clear_tv_win_move_timeout();

    tag_tv_win_move_timeout=gtk_timeout_add(200,tv_win_move_timeout,NULL);
  }

  return FALSE;
}

static gboolean cb_unmap()
{
//   dbg(".... unmap event\n");
   fftv_invisible();
   return FALSE;
}


static XF86VidModeModeInfo *disp_mode()
{
  static XF86VidModeModeInfo mod;
  int dotclock;


  XF86VidModeModeLine modl;
  XF86VidModeGetModeLine(dpy, screen_no,&dotclock,&modl);

  int modarrN;
  XF86VidModeModeInfo **modarr;
  XF86VidModeGetAllModeLines(dpy,screen_no,&modarrN, &modarr);

  int i;
  for(i=0; i<modarrN; i++) {
    XF86VidModeModeInfo *pmod = modarr[i];

    if (pmod->dotclock == dotclock &&
        pmod->hdisplay == modl.hdisplay &&  pmod->vdisplay == modl.vdisplay) {
       mod = *pmod;
       break;
    }
  }

  XFree(modarr);
  return &mod;
}

static XF86VidModeModeInfo mod_original;

void save_original_dpy_mode()
{
  mod_original = *disp_mode();
}

void restore_original_dpy_mode()
{
  XF86VidModeSwitchToMode(dpy, screen_no, &mod_original);
  XFlush(dpy);
}


void close_v4l();
void free_chanlist();
extern GtkWidget *tv_ui_window;
extern int b_do_exit;

void do_exit_fftv()
{
  dbg("......... do_exit_fftv ...\n");

  restore_original_dpy_mode();

  save_current_channel_name();

  tv_pic_off();
  tv_audio_off();
  close_v4l();

  dbg("......... do_exit_fftv b ...\n");
  free_chanlist();

  dbg("program will exit\n");
#if 0
  gtk_widget_destroy(gtv_win);
  gtk_widget_destroy(tv_ui_window);
#endif

  exit(0);
}


void fullscreen_toggle()
{
  int mxl = DisplayWidth(dpy,screen_no);
  int myl = DisplayHeight(dpy,screen_no);

  if (full_win640 && !full_win) {
    return;
  }

  tv_pic_off();

  if (!full_win) {
    owin_xl = win_xl;
    owin_yl = win_yl;
    owin_x = win_x;
    owin_y = win_y;
    win_xl = mxl;
    win_yl = myl;
    win_borderless=1;
    win_x=win_y=0;
    create_tv_win(FALSE, FALSE);
    full_win=1;
    move_mouse_xy(0,0);
  } else {
    win_xl = owin_xl;
    win_yl = owin_yl;
    win_x = owin_x;
    win_y = owin_y;
//    dbg("---------- %d %d\n",owin_xl,owin_yl);
    win_borderless=owin_borderless;
    create_tv_win(TRUE, FALSE);
    full_win=0;
  }

  update_pre_win_size();
}



void center_mouse_pointer()
{
  XF86VidModeSetViewPort(dpy, screen_no, win_x, win_y);
  move_mouse_xy(win_xl/2, win_yl/2);
}


void fullscreen_toggle640()
{
  int mxl = 640;
  int myl = 480;
  int i;
  static XF86VidModeModeInfo mod_last;

  tv_pic_off();

  if (!full_win640) {
    owin_xl = win_xl;
    owin_yl = win_yl;
    owin_x = win_x;
    owin_y = win_y;
    win_xl = mxl;
    win_yl = myl;
    win_borderless=1;
    win_x=win_y=150;

    create_tv_win(FALSE, FALSE);

    full_win640=1;
    mod_last = *disp_mode();

    int modarrN;
    XF86VidModeModeInfo **modarr;
    XF86VidModeGetAllModeLines(dpy,screen_no,&modarrN, &modarr);

    for(i=0;i<modarrN;i++) {
      XF86VidModeModeInfo *pmod = modarr[i];

      if (pmod->hdisplay==640 && pmod->vdisplay==480) {
        XF86VidModeSwitchToMode(dpy, screen_no, pmod);
#if 1
        XGrabPointer(dpy, tv_win, False,
             PointerMotionMask|ButtonPressMask,
             GrabModeAsync, GrabModeAsync, tv_win, None, CurrentTime);
        XGrabKeyboard(dpy, tv_win, False,
             GrabModeAsync, GrabModeAsync, CurrentTime);
#endif
        break;
      }
    }

    XFree(modarr);
  } else {
    win_xl = owin_xl;
    win_yl = owin_yl;
    win_x = owin_x;
    win_y = owin_y;
//    win_borderless=owin_borderless;
    win_borderless=0;

    create_tv_win(TRUE, FALSE);

    full_win640=0;

    XF86VidModeSwitchToMode(dpy, screen_no, &mod_last);
    XUngrabPointer(dpy, CurrentTime);
    XUngrabKeyboard(dpy, CurrentTime);
  }

  center_mouse_pointer();
}


int b_fftv_invisible;

void ExposeEve0()
{
  b_fftv_invisible=0;
#if 0
  dbg("--- ExposeEve0 %d %d\n",win_x,win_y);
#endif
  tv_pic_off();

  getTVWinXY(&win_x,&win_y);

  set_dpy_visibily_clips();

  if (tv_preview_mode) {
    add_tv_preview_timer(300);
  }

  return;
}


static int tag_tv_win_delayed_expose_timeout;
static gboolean tv_win_delayed_expose_timeout()
{
  tag_tv_win_delayed_expose_timeout =0;

  set_dpy_visibily_clips();

//  dbg("refresh_covering_window\n");
  refresh_covering_window();
#if 1
  if (!b_fftv_invisible)
    set_vwin();
#endif
  return FALSE;
}


static void tv_win_delayed_expose()
{
//  dbg("iiii\n");

  ExposeEve0();

  if (tag_tv_win_delayed_expose_timeout) {
    gtk_timeout_remove(tag_tv_win_delayed_expose_timeout);
    tag_tv_win_delayed_expose_timeout = 0;
  }

  tag_tv_win_delayed_expose_timeout =
     gtk_timeout_add(200, tv_win_delayed_expose_timeout, NULL);
}


gboolean cb_ExposeEve(GtkWidget *widget, GdkEventExpose *event, gpointer user_data)
{
  tv_win_delayed_expose();
#if 0
  dbg("cb_ExposeEve ...\n");
#endif
  return FALSE;
}


static gboolean update_win(Window win, int tvx0, int tvy0, int tvx1,int tvy1)
{
    XWindowAttributes watt;
    XEvent xeve;
    XExposeEvent *pxexpo= &xeve.xexpose;

    if (win == tv_win || win == frame_win)
      return FALSE;

    XGetWindowAttributes(dpy, win, &watt);

    if (!(watt.map_state & IsViewable))
      return FALSE;

    if (watt.override_redirect) {
      dbg("..... override_redirect\n");
      return FALSE;
    }
#if 1
    // mozilla is very weird
    if (!(watt.all_event_masks&(ExposureMask|VisibilityChangeMask))) {
//      dbg("no expose win\n");
      return FALSE;
    }
#endif

    pxexpo->width = watt.width;
    pxexpo->height = watt.height;
    pxexpo->x = 0;
    pxexpo->y = 0;

    pxexpo->type = Expose;
    pxexpo->window = win;
    pxexpo->send_event = True;
#if 1
    if (!XSendEvent(dpy,win,False,0, &xeve)) {
      dbg("send failed");
      return FALSE;
    }
#endif
//    XSync(dpy,False);
//    dbg("send %d %d   %d %d\n",pxexpo->x,pxexpo->y,pxexpo->width,pxexpo->height);
    return TRUE;
}

extern Window x_tv_ui_window;

void refreshTree(Window inwin, int tvx0, int tvy0, int tvx1,int tvy1, int depth,
                 gboolean my_own)
{
  if (depth > 100)
    return;

  int x0,y0,x1,y1;
  getGlobalWinRect(inwin,&x0,&y0,&x1,&y1);

  if (x0>tvx1 || y0>tvy1 || x1<tvx0 || y1<tvy0)
    return;

  Window r_root, r_parent, *child;
  int childn;

  XQueryTree(dpy, inwin, &r_root,&r_parent,&child,&childn);

  if (inwin==x_tv_ui_window)
    my_own = TRUE;

#if 1
  if (!update_win(inwin,tvx0,tvy0,tvx1,tvy1))
    return;
#else
  update_win(inwin,tvx0,tvy0,tvx1,tvy1);
#endif

  int i;
  for(i=0; i < childn; i++) {
    Window win=child[i];

    if (win == tv_win || win == frame_win)
      continue;

    refreshTree(win,tvx0,tvy0,tvx1,tvy1, depth + 1, False);
  }

  XFree(child);

}


static Window get_frame_win()
{
  Window r_root,r_parent;
  Window framewin;
  Window *child;
  int childn;

  r_parent=tv_win;
  do {
    framewin=r_parent;
    XQueryTree(dpy,framewin,&r_root,&r_parent,&child,&childn);
    XFree(child);
  } while (r_parent!=r_root);

  return framewin;
}


static void refresh_covering_window()
{
  Window r_root,r_parent;
  Window *child;
  int childn;
  int tvx0,tvy0,tvx1,tvy1,i;

  void *olderr = XSetErrorHandler(xerror_handler);

  // find the frame win that encloses tv_win

  XQueryTree(dpy,root,&r_root,&r_parent,&child,&childn);

  getGlobalWinRect(tv_win,&tvx0,&tvy0,&tvx1,&tvy1);
  win_x=tvx0; win_y=tvy0;

  for(i=0;i<childn && child[i]!=frame_win;i++);

  for(i++;i<childn;i++) {
    Window win=child[i];

    refreshTree(win,tvx0, tvy0, tvx1,tvy1, 0, False);
  }


  XFree(child);

  XSetErrorHandler(olderr);
}



#define DBGXEVE (0)

static gboolean VisibilityNotify_event(GtkWidget *widget, GdkEvent *event,
                                       gpointer user_data)
{
     if (!tv_win)
       return TRUE;

     tv_pic_off();
     GdkEventVisibility *peve=&event->visibility;
#if DBGXEVE
      dbg("VisibilityNotify: %d ",peve->state);
#endif
      if(peve->state == GDK_VISIBILITY_FULLY_OBSCURED) {
#if DBGXEVE
        dbg("FO\n");
#endif
        fftv_invisible();
      } else
      if(peve->state == GDK_VISIBILITY_UNOBSCURED) {
#if DBGXEVE
        dbg("U\n");
#endif
        clear_clips();
        ExposeEve0();
      } else {
#if DBGXEVE
        dbg("PO\n");
#endif
        tv_win_delayed_expose();
      }

      return TRUE;
}

static long last_reset_sreensaver_time;

void reset_sreensaver(char *caller)
{
   dbg(".. reset screensaver %s ..\n", caller);
   if (current_time() - last_reset_sreensaver_time < 60)
     return;

   last_reset_sreensaver_time = current_time();
   KeyCode kc_shift_l = XKeysymToKeycode(dpy, XK_Shift_L);
   XTestFakeKeyEvent(dpy, kc_shift_l, True, CurrentTime);
   XTestFakeKeyEvent(dpy, kc_shift_l, False, CurrentTime);
   XSync(dpy, False);
}

#if LIRC
void lirc_callback(gpointer data, gint source, GdkInputCondition condition)
{
  reset_sreensaver("lirc");

  lirc_fetch();
}
#endif


static jmp_buf disp_main_loop_jmp_buf;
static gboolean b_in_tv_disp_main_loop=FALSE;


void break_tv_disp_loop(gboolean b_clear_change_tv_win_refresh_mode)
{

  if (b_clear_change_tv_win_refresh_mode)
    b_change_tv_win_refresh_mode = FALSE;

  if (b_in_tv_disp_main_loop) {
    longjmp(disp_main_loop_jmp_buf, 1);
  }
  else {
    dbg("break_tv_disp_loop gtk_main_quit\n");
    gtk_main_quit();
  }
}


void refresh_local_win();

#define DISP_FRAME_RATE (30)
void tv_disp_main_loop()
{
retry:
  if ((fftv_mode & FFTV_MODE_TV) && tv_win_refresh_mode == TV_WIN_REFRESH_GRAB) {
    if (setjmp(disp_main_loop_jmp_buf)) {
      b_in_tv_disp_main_loop = FALSE;
      goto ret;
    }

    b_in_tv_disp_main_loop = TRUE;
    for(;;) {
      if (tv_preview_mode)
        refresh_local_win();
      else
        refresh_grab_tv_win();
#if 1
      usleep(1000000/DISP_FRAME_RATE + 1);
#else
      usleep(5000000/DISP_FRAME_RATE + 1);
#endif
      while (gtk_events_pending ())
         gtk_main_iteration ();
    }
  } else {
    gtk_main();
  }

  if (b_change_tv_win_refresh_mode) {
    b_change_tv_win_refresh_mode = FALSE;
    goto retry;
  }
ret:
  dbg("tv_disp_main_loop return\n");
}


void hide_tv_win()
{
  tv_pic_off();
  gtk_widget_hide(gtv_win);
  hide_tv_list_window();
  hide_tv_ui_window();
}

void process_gtk_events()
{
  while (gtk_events_pending ())
     gtk_main_iteration ();
}

void show_tv_ui_window(), show_tv_list_window();

void show_tv_ui_list_window()
{
  show_tv_ui_window();
  show_tv_list_window();
}
